<html>
  <head>
    <title>Stream</title>
  </head>
  <meta name="viewport" content="width=device-width,initial-scale=1.0,maximum-scale=1.0,minimum-scale=1.0,user-scalable=no">
  <body>
    <img style="max-width:100%;overflow:hidden;" src="{{ url_for('video_feed') }}">
  </body>
</html>
